# This file is placed in the Public Domain.


import inspect
import os
import queue
import threading
import time


from obj import Cfg, Object, get, keys, getname
from prs import parse, spl


def __dir__():
    return (
        "Bus",
        "Dispatcher",
        "Event",
        "Loop",
        "Table",
        "Thr",
        "clz",
        "typez",
        "launch"
    )


starttime = time.time()


class Bus(Object):

    objs = []

    @staticmethod
    def add(obj):
        Bus.objs.append(obj)

    @staticmethod
    def announce(txt):
        for h in Bus.objs:
            if "announce" in dir(h):
                h.announce(txt)

    @staticmethod
    def byorig(orig):
        for o in Bus.objs:
            if repr(o) == orig:
                return o

    @staticmethod
    def first(otype=None):
        if Bus.objs:
            if not otype:
                return Bus.objs[0]
            for o in Bus.objs:
                if otype in str(type(o)):
                    return o
        return None

    @staticmethod
    def say(orig, channel, txt):
        for o in Bus.objs:
            if repr(o) == orig:
                o.say(channel, txt)

class Dispatcher(Object):

    def __init__(self):
        super().__init__()
        self.cbs = Object()

    def dispatch(self, event):
        if event and event.type in self.cbs:
            self.cbs[event.type](self, event)
        else:
            event.ready()

    def register(self, k, v):
        self.cbs[str(k)] = v


class Event(Object):

    def __init__(self):
        super().__init__()
        self.channel = None
        self.done = threading.Event()
        self.orig = None
        self.result = []
        self.type = "event"
        self.txt = None

    def bot(self):
        return Bus.byorig(self.orig)

    def parse(self):
        parse(self, self.txt)

    def ready(self):
        self.done.set()

    def reply(self, txt):
        self.result.append(txt)

    def show(self):
        for txt in self.result:
            Bus.say(self.orig, self.channel, txt)
        self.ready()

    def wait(self, timeout=1.0):
        self.done.wait(timeout)


class Loop(Dispatcher):

    def __init__(self):
        super().__init__()
        self.queue = queue.Queue()
        self.speed = "normal"
        self.stopped = threading.Event()
        self.register("cmd", self.handle)

    def do(self, e):
        self.dispatch(e)

    def event(self, txt):
        e = Event()
        e.type = "cmd"
        e.txt = txt
        e.orig = repr(self)
        return e

    def handle(self, clt, e):
        e.parse()
        f = get(Table.cmds, e.cmd, None)
        if f:
            f(e)
            e.show()
        e.ready()

    def loop(self):
        self.stopped.clear()
        while not self.stopped.isSet():
            self.do(self.event(self.poll()))

    def poll(self):
        return self.queue.get()

    def put(self, e):
        self.queue.put_nowait(e)

    def restart(self):
        self.stop()
        self.start()

    def start(self):
        launch(self.loop)
        return self

    def stop(self):
        self.stopped.set()
        self.queue.put(None)


class Table(Object):

    classes = Object()
    cmds = Object()
    modules = Object()
    names = Object()

    @staticmethod
    def add(mod):
        Table.modules[mod.__name__] = mod

    @staticmethod
    def get(mn):
        return get(Table.modules, mn, None)

    @staticmethod
    def introspect(mod):
        if not mod:
            return
        if mod not in Table.modules:
            Table.modules[mod.__name__.lower()] = mod
        for k, o in inspect.getmembers(mod, inspect.isfunction):
            if k not in Table.cmds:
                if o.__code__.co_argcount == 1:
                    if "event" in o.__code__.co_varnames:
                        Table.cmds[k] = o
        for k, clz in inspect.getmembers(mod, inspect.isclass):
            if k not in Table.classes and issubclass(clz, Object):
                Table.classes[clz.__name__.lower()] = clz
                name = clz.__name__.lower()
                if name not in Table.names:
                    Table.names[name] = []
                clzname = "%s.%s" % (clz.__module__, clz.__name__)
                if clzname not in Table.names[name]:
                    Table.names[name].append(clzname)

    @staticmethod
    def scan(skip=None):
        ml = ",".join(keys(Table.modules))
        for mn in spl(ml):
            if skip and mn in spl(skip):
                continue
            Table.introspect(Table.get(mn))



class Thr(threading.Thread):

    def __init__(self, func, *args, thrname="", daemon=True):
        super().__init__(None, self.run, thrname, (), {}, daemon=daemon)
        self.name = thrname or getname(func)
        self.result = None
        self.queue = queue.Queue()
        self.queue.put_nowait((func, args))
        self.sleep = 0

    def __iter__(self):
        return self

    def __next__(self):
        for key in dir(self):
            yield key

    def join(self, timeout=None):
        super().join(timeout)
        return self.result

    def run(self):
        func, args = self.queue.get_nowait()
        if args:
            target = vars(args[0])
            if target and "txt" in dir(target):
                self.name = target.txt.split()[0]
        self.setName(self.name)
        self.result = func(*args)


def launch(func, *args, **kwargs):
    name = kwargs.get("name", getname(func))
    t = Thr(func, *args, thrname=name, daemon=True)
    t.start()
    return t
